package client;

import core.LufBoard;
import core.Position;
import core.TileMove;
import core.TileType;
import java.io.IOException;
import java.net.UnknownHostException;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import network.*;
import tools.TileTools;

/**

 @author reidl
 */
public class LufClient implements Runnable, NetworkController.Listener {

    private static final boolean PRINT_DEBUG = false;
    private Random rnd;
    private boolean isRunning = false;
    private int knockPort;
    private String name;
    private NetworkController controller = null;
    private int cliendId = -1;
    private Set<Integer> objectives = null;
    private LufBoard board = null;

    public LufClient(String name, int knockPort) {
        this.name = name;
        this.knockPort = knockPort;
        this.rnd = new Random();
    }

    public void start() {
        assert (!isRunning);
        isRunning = true;
        System.out.println("Starting Client '" + name + "'");
        (new Thread(this)).start();
    }

    private void debug(String msg) {
        if (!PRINT_DEBUG) {
            return;
        }

        System.out.print(name);
        System.out.print(": ");
        System.out.println(msg);
    }

    @Override
    public void run() {
        try {
            // Try to connect
            boolean connected = false;
            while (isRunning && !connected) {
                try {
                    controller = NetworkController.connect("localhost", knockPort, name);

                    if (controller == null) {
                        // Something awful happened (protocol violated)
                        stop();
                    } else {
                        connected = true;
                    }
                } catch (java.net.ConnectException ex) {
                    // Server to yet ready
                }
            }

            // Message handling loop
            while (isRunning) {
                controller.readMessages(this);
            }
        } catch (UnknownHostException ex) {
            Logger.getLogger(LufClient.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(LufClient.class.getName()).log(Level.SEVERE, null, ex);
        } catch (MalformedMessageException ex) {
            Logger.getLogger(LufClient.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public void handleMessage(MessageGameStart msg) throws IOException {
        board = msg.createBoard();
        cliendId = msg.getPlayerId();
        objectives = new TreeSet<Integer>();
        objectives.addAll(msg.getPlayerObjectives());
        debug("Got game start message, my objectives are (" + objectives.size() + "): " + objectives);
    }

    @Override
    public void handleMessage(MessageGameEnd msg) throws IOException {
        if (msg.getWinners().contains(cliendId)) {
            debug("Game ended. I won.");
        } else {
            debug("Game ended. I lost.");
        }
        stop();
    }

    @Override
    public void handleMessage(MessageYourTurn msg) throws IOException {
        debug("My turn! I currently sit at " + board.getPlayerPosition(cliendId));

        // Do some random move
        TileMove move = randomMove();
        LufBoard testBoard = board.copy();
        testBoard.insertTile(move);

        Position currentPosition = testBoard.getPlayerPosition(cliendId);
        Position targetPosition = null;

        // Try to collect something
        for (Integer objective : objectives) {
            Position objPosition = board.getObjectivePosition(objective);
            // TODO: should not be null. check why it is sometimes.
            if (objPosition != null && testBoard.isReachable(currentPosition, objPosition)) {
                targetPosition = objPosition;
                debug("Yeah! I can reach objective " + objective + " at " + objPosition);
                break;
            }
        }

        // Alternative: just go somewhere
        if (targetPosition == null) {
            do {
                targetPosition = Position.create(rnd.nextInt(board.getWidth()), rnd.nextInt(board.getHeight()));
            } while (!testBoard.isReachable(currentPosition, targetPosition));
        }

        debug("Inserting tile " + move + " and moving from " + currentPosition + " to " + targetPosition);

        controller.announceMove(move, board.getFreeTile().getType(), cliendId, targetPosition);
    }

    @Override
    public void handleMessage(MessageKickPlayer msg) throws IOException {
        if (msg.playerId == cliendId) {
            debug("Oh shit, I was kicked!");
            stop();
            return;
        }

        debug("Player " + msg.playerId + " was kicked.");
        board.removePlayer(msg.playerId);
    }

    @Override
    public void handleMessage(MessageMove msg) throws IOException {
        board.insertTile(msg.getTileMove());
        board.movePlayer(msg.getPlayer(), msg.getPlayerMove());

        for (Map.Entry<Integer, Set<Integer>> entry : msg.getCollectionEvents().entrySet()) {
            int playerId = entry.getKey();
            Set<Integer> collected = entry.getValue();
            board.collectObjectives(playerId, collected);

            if (playerId == cliendId) {
                assert (objectives.containsAll(collected));
                objectives.removeAll(collected);
                debug("I collected objectives " + collected + ", only " + objectives.size() + " remain.");
            }
        }
    }

    @Override
    public void handleMessage(MessagePlayerInfo msg) throws IOException {
        // Not needed for client.
    }

    @Override
    public void fail() {
        stop();
    }

    private void stop() {
        isRunning = false;
    }

    private TileMove randomMove() {
        TileMove move = null;
        TileType freeTile = board.getFreeTile().getType();

        // Rotate arbitrarily
        TileType[] trace = TileTools.getTrace(freeTile);
        freeTile = trace[rnd.nextInt(trace.length)];

        do {
            int width = board.getWidth();
            int height = board.getHeight();
            int dir = rnd.nextInt(4);
            if (dir == 1) {
                move = TileMove.left(rnd.nextInt(width), freeTile);
            } else if (dir == 2) {
                move = TileMove.right(rnd.nextInt(width), freeTile);
            } else if (dir == 3) {
                move = TileMove.top(rnd.nextInt(height), freeTile);
            } else {
                move = TileMove.bottom(rnd.nextInt(height), freeTile);
            }
        } while (!board.isMoveLegal(move));
        return move;
    }
}
