package core;

import exceptions.IllegalMoveException;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import network.*;
import setup.BoardBuilder;
import tools.Iterators;

/**
 *
 * @author reidl
 */
public class GameServer implements Runnable {

    private boolean slowmo = false;
    private LufBoard board = null;
    private boolean isRunning = false;
    private boolean isPlaying = false;
    private BoardBuilder boardFactory;
    private int timeout = 1000;
    private int knockPort = -1;
    private int numPlayers = 0;
    private int currentPlayer = 0;
    private ArrayList<Player> players = new ArrayList<Player>();
    private ArrayList<Spectator> spectators = new ArrayList<Spectator>();
    private GameServerLog broadcastLog = new GameServerLog();
    private int delay;

    public GameServer() {
    }

    public GameServer setBoardFactory(BoardBuilder boardFactory) {
        this.boardFactory = boardFactory;
        return this;
    }

    public GameServer setPort(int port) {
        this.knockPort = port;
        return this;
    }

    public GameServer setLogfile(String path) {
        broadcastLog.setLogFile(path);
        return this;
    }

    public GameServer setNumPlayers(int numPlayers) {
        // Game starts as soon as this many players have connected
        // 0 denotes an arbitrary numbers (start after timeout?)
        this.numPlayers = numPlayers;
        return this;
    }
    
    public GameServer setDelay(int delay) {
	if (delay > 0) {
	    this.delay = delay;
	    this.slowmo = true;
	}
	return this;
    }

    public void start() {
        assert (!isRunning);
        isRunning = true;
        players.clear();
        System.out.println("Serv: Starting game server...");
        KnockServer.create(this, knockPort);
    }

    @Override
    public void run() {
        // Give clients some time
        try {
            synchronized (this) {
                wait(1000);
            }
        } catch (InterruptedException ex) {
            Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
        }

        System.out.println();
        System.out.println("Serv: The game is on!");
        System.out.println();

        // Main game loop
        boolean gameFinished = false;
        while (!gameFinished) {
            if (slowmo) {
                try {
                    synchronized (this) {
                        wait(this.delay);
                    }
                } catch (InterruptedException ex) {
                    Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
                }
            }

            // Debug output
            System.out.print("Serv: Player positions are ");
            for (Player player : players) {
                Position pos = board.getPlayerPosition(player.id);
                System.out.print(player.toString() + " " + pos + ", ");
            }
            System.out.println();

            Player movingPlayer = players.get(currentPlayer);

            MessageYourTurn msgYourTurn = new MessageYourTurn();
            msgYourTurn.timeout = timeout;
            System.out.println("Serv: player " + movingPlayer);

            MessageMove msgMove = null;
            LufBoard boardCopy = board.copy();
            try {
                // Send client message
                msgYourTurn.write(movingPlayer.getStream());
                movingPlayer.setTimeout((int) msgYourTurn.timeout);
                msgMove = movingPlayer.readMoveMessage();

                System.out.println("Serv: Got move message " + msgMove);

                board.insertTile(msgMove.getTileMove());
                board.movePlayer(movingPlayer.id, msgMove.getPlayerMove());
            } catch (IllegalMoveException ex) {
                kickPlayer(movingPlayer, NetworkConstants.REASON_ILLEGAL_MOVE);
                board = boardCopy; // Reset move
                Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
            } catch (MalformedMessageException ex) {
                kickPlayer(movingPlayer, NetworkConstants.REASON_MALFORMED_MESSAGE);
                Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
            } catch (SocketTimeoutException ex) {
                kickPlayer(movingPlayer, NetworkConstants.REASON_MOVE_TIMEOUT);
                Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IOException ex) {
                kickPlayer(movingPlayer, NetworkConstants.REASON_NETWORK_ERROR);
                Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
            }

            ArrayList<Integer> winners = new ArrayList<Integer>();
            if (msgMove != null) {
                System.out.println("Serv: move " + msgMove.getTileMove() + ", " + msgMove.getPlayerMove());

                // Augment message with collection events
                msgMove.clearCollectionEvents();
                for (Player player : players) {
                    Set<Integer> collected = board.collectObjectives(player.id);
                    if (!collected.isEmpty()) {
                        System.out.println("Serv: player " + player.name + " collected objectives " + collected);
                        msgMove.addCollectionEvents(player.id, collected);
                        if (board.getPlayerObjectives(player.id).isEmpty()) {
                            winners.add(player.id);
                        }
                    }
                }

                broadCast(msgMove);
            }

            if (numPlayers == 0 || !winners.isEmpty()) {
                MessageGameEnd msgGameEnd = new MessageGameEnd();
                msgGameEnd.setWinners(winners);
                broadCast(msgGameEnd);
                gameFinished = true;
            } else {
                currentPlayer = (currentPlayer + 1) % numPlayers;
            }
        }

        System.out.println("Serv: Game ended.");
        System.exit(0); // For now.
    }

    public void acceptClient(String name, ServerSocket socket) {
        if (isPlaying) {
            return;
        }

        System.out.println("Serv: Client " + name + " connected to " + socket.getLocalPort());
        try {
            players.add(new Player(name, players.size(), socket.accept()));
        } catch (IOException ex) {
            Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
        }

        if (players.size() == numPlayers) {
            startGame();
        } else {
            System.out.println("Serv: Got " + players.size() + " of " + numPlayers + " players");
        }
    }

    public void acceptSpectator(ServerSocket socket) {
        System.out.println("Serv: Spectator connected to " + socket.getLocalPort());
        try {
            Spectator spectator = new Spectator(socket.accept());
            spectators.add(spectator);
            broadcastLog.replay(spectator);
        } catch (IOException ex) {
            Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private void kickPlayer(Player player, byte reason) {
        System.out.println("Serv: Kicking player " + player);
        
        // First broadcast so we know even the kicked player
        // gets the message
        MessageKickPlayer msgKickPlayer = new MessageKickPlayer();
        msgKickPlayer.playerId = player.id;
        msgKickPlayer.reason = reason;
        broadCast(msgKickPlayer);

        // Remove player
        int index = players.indexOf(player);
        assert (index != -1);

        players.remove(player);
        numPlayers = players.size();

        if (currentPlayer > index) {
            currentPlayer--;
        }
    }

    private void broadCast(Message message) {
        broadcastLog.logMessage(message);

        HashSet<Spectator> remove = new HashSet<Spectator>();
        for (Spectator spectator : Iterators.join(players, spectators)) {
            try {
                message.write(spectator.getStream());
            } catch (IOException ex) {
                remove.add(spectator);
                Logger.getLogger(GameServer.class.getName())
                        .log(Level.SEVERE, "Failed for " + spectator, ex);
            }
        }
        
        // Spectators are just removed
        spectators.removeAll(remove);
        
        // TODO: kick players.
        
    }

    private void startGame() {
        // Mix up player order
        Collections.shuffle(players);

        // Create board 
        board = boardFactory.setNumPlayers(players.size()).build();

        MessageGameStart msgGameStart = MessageGameStart.fromBoard(board);
        MessagePlayerInfo msgPlayerInfo = new MessagePlayerInfo(numPlayers);
        for (Player player : players) {
            msgPlayerInfo.setPlayerName(player.id, player.name);
        }
        
        System.out.println("Serv: broadcasting game start message");

        for (Spectator spectator : spectators) {
            try {
                // TODO: Write in specs
                msgPlayerInfo.write(spectator.getStream());
                msgGameStart.write(spectator.getStream());                
            } catch (IOException ex) {
                Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
            }
        }

        broadcastLog.logMessage(msgPlayerInfo);
        broadcastLog.logMessage(msgGameStart);        

        // Write private part of messages and submit to players
        for (Player player : players) {
            msgGameStart.setPlayerId(player.id);
            msgGameStart.setPlayerObjectives(board.getPlayerObjectives(player.id));
            try {
                System.out.println("Sending game start message to " + player.name);
                msgGameStart.write(player.getStream());
            } catch (IOException ex) {
                kickPlayer(player, NetworkConstants.REASON_NETWORK_ERROR);
                Logger.getLogger(GameServer.class.getName()).log(Level.SEVERE, null, ex);
            }
        }

        // Start main loop
        isPlaying = true;
        (new Thread(this)).start();
    }

    public GameServer setTimeout(int t) {
        this.timeout = t;
        return this;
    }

    protected static class Player extends Spectator {

        public final String name;
        private int id;

        public Player(String name, int id, Socket socket) {
            super(socket);
            this.name = name;
            this.id = id;
        }

        public MessageMove readMoveMessage() throws IOException, MalformedMessageException {
            DataInputStream in = new DataInputStream(socket.getInputStream());
            byte header = in.readByte();
            if (header != NetworkConstants.MSG_MOVE) {
                throw new MalformedMessageException();
            }
            return MessageMove.parse(in);
        }

        public void setTimeout(int millis) throws SocketException {
            socket.setSoTimeout(millis);
        }

        public String toString() {
            return name + "(" + id + ")";
        }
    }

    protected static class Spectator {

        protected final Socket socket;

        public Spectator(Socket socket) {
            this.socket = socket;
        }

        public DataOutputStream getStream() throws IOException {
            if(socket.isClosed())
                System.err.println("Socket was closed for " + toString());
            return new DataOutputStream(socket.getOutputStream());
        }

        public String toString() {
            return "(spectator)";
        }
    }
}
