package core;

/**
 *
 * @author ubik
 */
public abstract class LufBoardEvent {
    
    public abstract void accept(Visitor visitor);
    
    public interface Visitor {
        public void visit(NewBoard newBoardEvent);
        public void visit(PlayerData playerDataEvent);
        public void visit(NewTile newTileEvent);
        public void visit(InsertFreeTile insertFreeTileEvent);
        public void visit(MakeTileFree makeTileFreeEvent);
        public void visit(MoveTile moveTileEvent);
        public void visit(MovePlayer movePlayerEvent);
        public void visit(KillPlayer killPlayerEvent);
    }
    
    public static class NewBoard extends LufBoardEvent {
        public final int width, height;

        public NewBoard(int width, int height) {
            this.width = width;
            this.height = height;
        }

        @Override
        public void accept(Visitor visitor) {
            visitor.visit(this);
        }

        @Override
        public String toString() {
            return "NewBoard{" + "width=" + width + ", height=" + height + '}';
        }
    }
    
    public static class PlayerData extends LufBoardEvent {
        public final String[] names;

        public PlayerData(String[] names) {
            this.names = names;
        }

        @Override
        public void accept(Visitor visitor) {
            visitor.visit(this);
        }

        @Override
        public String toString() {
            return "PlayerData{" + "names=" + names + '}';
        }
    }
    
    public static class NewTile extends LufBoardEvent {
        public final int id;
        public final int x,y;
        public final TileType type;
        public final int[] objectives;
        public final int[] players;
        public final boolean isFree, isMovable;  

        public NewTile(int id, int x, int y, int[] objectives, int[] players, TileType type, boolean isFree, boolean isMovable) {
            this.id = id;
            this.x = x;
            this.y = y;
            this.objectives = objectives;
            this.players = players;
            this.type = type;
            this.isFree = isFree;
            this.isMovable = isMovable;
        }
        
        @Override
        public void accept(Visitor visitor) {
            visitor.visit(this);
        }

        @Override
        public String toString() {
            return "NewTile{" + "id=" + id + ", x=" + x + ", y=" + y + ", type=" + type + ", objectives=" + objectives + ", players=" + players + ", isFree=" + isFree + ", isMovable=" + isMovable + '}';
        }
    }
    
    public static class MoveTile extends LufBoardEvent {
        public final int id;
        public final Position pos;

        public MoveTile(int id, Position pos) {
            this.id = id;
            this.pos = pos.copy();
        }

        @Override
        public void accept(Visitor visitor) {
            visitor.visit(this);
        }

        @Override
        public String toString() {
            return "MoveTile{" + "id=" + id + ", pos=" + pos + '}';
        }
    }
    
    public static class InsertFreeTile extends LufBoardEvent {
        public final int id;
        public final int rotations; // Number of CW 90deg rotations
        public final Position pos;
        public final BoardEdge edge;

        public InsertFreeTile(int id, int rotations, Position pos, BoardEdge edge) {
            this.id = id;
            this.rotations = rotations;
            this.pos = pos.copy();
            this.edge = edge;
        }
        
        @Override
        public void accept(Visitor visitor) {
            visitor.visit(this);
        }

        @Override
        public String toString() {
            return "InsertFreeTile{" + "id=" + id + ", rotations=" + rotations + ", pos=" + pos + ", edge=" + edge + '}';
        }
    }
    
    public static class MakeTileFree extends LufBoardEvent {
        public final int id;
        public final Position pos; // Position of tile
        public final BoardEdge edge; // Edge of the exit

        public MakeTileFree(int id, Position pos, BoardEdge edge) {
            this.id = id;
            this.pos = pos.copy();
            this.edge = edge;
        }
        
        @Override
        public void accept(Visitor visitor) {
            visitor.visit(this);
        }

        @Override
        public String toString() {
            return "MakeTileFree{" + "id=" + id + ", pos=" + pos + ", edge=" + edge + '}';
        }
    }
    
    public static class MovePlayer extends LufBoardEvent {
        public final int id;
        public final int tileId;

        public MovePlayer(int id, int tileId) {
            this.id = id;
            this.tileId = tileId;
        }
        
        @Override
        public void accept(Visitor visitor) {
            visitor.visit(this);
        }

        @Override
        public String toString() {
            return "MovePlayer{" + "id=" + id + ", tileId=" + tileId + '}';
        }

    }
    
    public static class KillPlayer extends LufBoardEvent {
        public final int id;

        public KillPlayer(int id) {
            this.id = id;
        }
        
        @Override
        public void accept(Visitor visitor) {
            visitor.visit(this);
        }

        @Override
        public String toString() {
            return "KillPlayer{" + "id=" + id + '}';
        }
    }
}