package core;

import client.LufClient;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import org.apache.commons.cli.*;
import setup.BoardBuilder;
import setup.RandomBoardBuilder;

public class Main {

	private static Options options;

	private static void initializeOptions() {
		// create Options object
		Main.options = new Options();

		// add options
		Option knockPortOption = new Option("p", "knockPort", true, "Port for KnockServer to listen on");
		Option playersOption = new Option("n", "players", true, "Number of Players to wait for");
		Option timeoutOption = new Option("t", "timeout", true, "Timeout for each turn in msecs (server only)");
		Option boardSizeOption = new Option("s", "boardSize", true,
				"Size of the board in format width*height, eg. 10*10 (server only)");
		Option delayOption = new Option("d", "delay", true, "Delay between each turn (server only)");
		Option logFileOption = new Option("l", "logFile", true, "File to write logs to (server only)");
		Option clientOption = new Option("c", "client", false, "Start in client mode, not server");
		Option nameOption = new Option("a", "name", true, "Client Player Name (client only)");
		Option demoOption = new Option("d", "demo", false, "Start in demo mode, not server");

		Main.options.addOption(knockPortOption);
		Main.options.addOption(playersOption);
		Main.options.addOption(timeoutOption);
		Main.options.addOption(boardSizeOption);
		Main.options.addOption(delayOption);
		Main.options.addOption(logFileOption);
		Main.options.addOption(clientOption);
		Main.options.addOption(nameOption);
		Main.options.addOption(demoOption);
	}

	public static void main(String[] args) throws IOException {

		Main.initializeOptions();

		CommandLineParser clp = new GnuParser();

		try {
			// parse the command line arguments
			CommandLine line = clp.parse(Main.options, args);

			// Branch client <-> server <-> demo
			if (line.hasOption("client")) {
				// Client mode
				System.out.println("Starting in Client Mode.");
				// the client requires knockPort and name
				int knockPort = Main.parseIntOption(false, "knockPort", line, -1);
				String name = null;

				name = Main.parseStringOption(false, "name", line, null);

				new LufClient(name, knockPort).start();
			} else if (line.hasOption("demo")) {
				// Demo mode
				int port = 5555;
				int numClients = 4;
				GameServer server = new GameServer();
				BoardBuilder builder = new RandomBoardBuilder().setWidth(8).setHeight(8);

				String logFile = Main.parseStringOption(true, "logFile", line, null);

				if (logFile != null) {
					server.setLogfile(logFile);
				}

				server.setBoardFactory(builder).setPort(port).setNumPlayers(numClients).setTimeout(1000).start();

				new LufClient("Olive", port).start();
				new LufClient("Magenta", port).start();
				new LufClient("Teal", port).start();
				new LufClient("Raspberry", port).start();
			} else {
				// Server mode
				System.out.println("Starting in Server Mode.");

				// (" --knockPort=1234 -n=2 --boardSize=25*35 --timeout=5000");
				int knockPort;
				int numPlayers;
				int boardHeight = 10;
				int boardWidth = 10;
				int timeout = 1000;
				int delay = -1;
				String logFile = null;

				knockPort = Main.parseIntOption(false, "knockPort", line, -1);
				numPlayers = Main.parseIntOption(false, "players", line, -1);

				timeout = Main.parseIntOption(true, "timeout", line, 1000);
				delay = Main.parseIntOption(true, "delay", line, -1);

				logFile = Main.parseStringOption(true, "logFile", line, null);

				if (line.hasOption("boardSize")) {
					String boardSize = line.getOptionValue("boardSize");
					try {
						Pattern regex = Pattern.compile("(\\d)+\\*(\\d)+", Pattern.CASE_INSENSITIVE
								| Pattern.UNICODE_CASE);
						Matcher regexMatcher = regex.matcher(boardSize);
						if (regexMatcher.find()) {
							String width = regexMatcher.group(1);
							String height = regexMatcher.group(2);

							try {
								boardWidth = Integer.parseInt(width);
								boardHeight = Integer.parseInt(height);
							} catch (NumberFormatException exp) {
								System.out
										.println("Incorrect parameter string for boardSize: Must be in format width*height");
								Main.printHelpAndQuit();
							}
						} else {
							System.out
									.println("Incorrect parameter string for boardSize: Must be in format width*height");
							Main.printHelpAndQuit();
						}
					} catch (PatternSyntaxException ex) {
						// Syntax error in the regular expression
						System.out.println("Critical internal error.");
						Main.printHelpAndQuit();
					}
				}

				GameServer server = new GameServer();
				BoardBuilder builder = new RandomBoardBuilder().setWidth(boardWidth).setHeight(boardHeight);

				if (logFile != null) {
					server.setLogfile(logFile);
				}

				server.setBoardFactory(builder).setPort(knockPort).setNumPlayers(numPlayers).setTimeout(timeout)
						.setDelay(delay).start();
			}
		} catch (ParseException exp) {
			System.out.println("Parsing arguments failed: " + exp.getMessage());
			Main.printHelpAndQuit();
		} catch (NumberFormatException ex) {
			System.out.println("Parsing arguments failed (not a number): " + ex.getMessage());
			Main.printHelpAndQuit();
		}
	}

	private static int parseIntOption(boolean isOptional, String name, CommandLine line, int defaultValue) {
		int result = defaultValue;
		if (line.hasOption(name)) {
			try {
				result = Integer.parseInt(line.getOptionValue(name));
			} catch (NumberFormatException exp) {
				System.out.println("Parameter " + name + " was not a number!");
				Main.printHelpAndQuit();
			}
		} else if (!isOptional) {
			System.out.println("Required parameter " + name + " was missing.");
			Main.printHelpAndQuit();
		}

		return result;
	}

	private static String parseStringOption(boolean isOptional, String name, CommandLine line, String defaultValue) {
		String result = defaultValue;
		if (line.hasOption(name)) {
			result = line.getOptionValue(name);
		} else if (!isOptional) {
			System.out.println("Required parameter " + name + " was missing.");
			Main.printHelpAndQuit();
		}

		return result;
	}

	private static void printHelpAndQuit() {
		System.out.println("Game server & test client v1.1. Usage: ");

		HelpFormatter helpFormatter = new HelpFormatter();
		helpFormatter.printHelp("java -jar labyrinth.jar", Main.options);

		System.out.println("Server example cmd:");
		System.out.println(" --knockPort=1234 -n=2 --boardSize=25*35 --timeout=5000");
		System.out.println("Client example cmd:");
		System.out.println(" --client --knockPort=1234 --name=blub");
		System.out.println("Demo example cmd:");
		System.out.println(" --logFile=file.ext");

		System.out.println("Terminating. Goodbye...");
		System.exit(-1);
	}
}
